{% extends "base.html" %}

{% block title %}Debug messages{% endblock %}

{% block content %}

<h1>Debug messages</h1>

<p>Set a message:</p>

<form action="{{ urls.path('-/messages') }}" method="post">
    <div>
        <input type="text" name="message" style="width: 40%">
        <div class="select-wrapper">
            <select name="message_type">
                <option>INFO</option>
                <option>WARNING</option>
                <option>ERROR</option>
                <option>all</option>
            </select>
        </div>
        <input type="hidden" name="csrftoken" value="{{ csrftoken() }}">
        <input type="submit" value="Add message">
    </div>
</form>

{% endblock %}
