<script src="{{ base_url }}-/static/sql-formatter-2.3.3.min.js" defer></script>
<script src="{{ base_url }}-/static/cm-editor-6.0.1.bundle.js"></script>
<style>
  .cm-editor {
    resize: both;
    overflow: hidden;
    width: 80%;
    border: 1px solid #ddd;
  }
  /* Fix autocomplete icon positioning. The icon element gets border-box sizing set due to
     the global reset, but this causes overlapping icon and text. Markup:
     `<div class="cm-completionIcon cm-completionIcon-keyword" aria-hidden="true"></div>` */
  .cm-completionIcon {
    box-sizing: content-box;
  }
</style>
