<!-- above-table-panel is a hook node for plugins to attach to . Displays even if no data available -->
<div class="above-table-panel"> </div>
{% if display_rows %}
<div class="table-wrapper">
    <table class="rows-and-columns">
        <thead>
            <tr>
                {% for column in display_columns %}
                    <th {% if column.description %}data-column-description="{{ column.description }}" {% endif %}class="col-{{ column.name|to_css_class }}" scope="col" data-column="{{ column.name }}" data-column-type="{{ column.type.lower() }}" data-column-not-null="{{ column.notnull }}" data-is-pk="{% if column.is_pk %}1{% else %}0{% endif %}">
                        {% if not column.sortable %}
                            {{ column.name }}
                        {% else %}
                            {% if column.name == sort %}
                                <a href="{{ fix_path(path_with_replaced_args(request, {'_sort_desc': column.name, '_sort': None, '_next': None})) }}" rel="nofollow">{{ column.name }}&nbsp;▼</a>
                            {% else %}
                                <a href="{{ fix_path(path_with_replaced_args(request, {'_sort': column.name, '_sort_desc': None, '_next': None})) }}" rel="nofollow">{{ column.name }}{% if column.name == sort_desc %}&nbsp;▲{% endif %}</a>
                            {% endif %}
                        {% endif %}
                    </th>
                {% endfor %}
            </tr>
        </thead>
        <tbody>
        {% for row in display_rows %}
            <tr>
                {% for cell in row %}
                    <td class="col-{{ cell.column|to_css_class }} type-{{ cell.value_type }}">{{ cell.value }}</td>
                {% endfor %}
            </tr>
        {% endfor %}
        </tbody>
    </table>
</div>
{% else %}
    <p class="zero-results">0 records</p>
{% endif %}
